/* Breakpoint handling */
#include <config.h>

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

#include <breakpoint.h>
#include <expr.h>
#include <lib.h>

/* Breakpoint, including private data */
struct bp
{
  struct breakpoint v;
  struct bp **pprev;
  int prev_value; /* Breakpoint triggers only when its previous value is 0 */
};

/* Breakpoint list. Should be modified only the breakpoint_* functions */
struct breakpoint *breakpoint_list; /* = NULL; */
static struct bp **breakpoint_list_tail = (struct bp **)&breakpoint_list;

/* Create a new breakpoint with EXPR;
   returns: new breakpoint or NULL on error */
struct breakpoint *
breakpoint_new (const char *expr)
{
  intmax_t val;
  struct bp *b;
  struct expr_code *code;

  code = expr_compile (expr, 0);
  /* Check the expression is usable */
  if (code == NULL || expr_exec (code, &val) != 0)
    return NULL;
  b = xmalloc (sizeof (*b));
  b->v.expr = xstrdup (expr);
  b->v.compiled = code;
  b->v.next = NULL;
  b->pprev = breakpoint_list_tail;
  b->prev_value = 0;
  *breakpoint_list_tail = b;
  breakpoint_list_tail = (struct bp **)&b->v.next;
  return &b->v;
}

/* Change BP to EXPR;
   returns: 0 if OK, -1 on error */
int
breakpoint_change (struct breakpoint *bp, const char *expr)
{
  intmax_t val;
  struct expr_code *code;

  code = expr_compile (expr, 0);
  if (code == NULL || expr_exec (code, &val) != 0)
    return -1;
  free (bp->expr);
  free (bp->compiled);
  bp->expr = xstrdup (expr);
  bp->compiled = code;
  return 0;
}

/* Delete BP */
void
breakpoint_delete (struct breakpoint *bp)
{
  struct bp *b;

  assert (offsetof (struct bp, v) == 0); /* Guaranteed by the C standard */
  b = (struct bp *)bp;
  *b->pprev = (struct bp *)b->v.next;
  if (b->v.next != NULL)
    ((struct bp *)b->v.next)->pprev = b->pprev;
  if (breakpoint_list_tail == (struct bp **)&b->v.next)
    breakpoint_list_tail = b->pprev;
  free (b->v.expr);
  free (b->v.compiled);
  free (b);
}

/* Check whether BP triggered */
int
breakpoint_eval (struct breakpoint *bp)
{
  intmax_t val;

  return expr_exec (bp->compiled, &val) == 0 && val != 0;
}

/* Check whether some breakpoint triggered;
   returns: first matching breakpoint or NULL if none */
struct breakpoint *
breakpoint_check (void)
{
  struct breakpoint *bp, *res;

  res = NULL;
  for (bp = breakpoint_list; bp != NULL; bp = bp->next)
    {
      struct bp *b;
      int v;

      assert (offsetof (struct bp, v) == 0); /* Guaranteed by the C standard */
      b = (struct bp *)bp;
      v = breakpoint_eval (bp);
      if (res == NULL && v != 0 && b->prev_value == 0)
	res = bp;
      b->prev_value = v;
    }
  return res;
}

/* Delete all known breakpoints */
void
breakpoint_delete_all (void)
{
  struct bp *b;

  assert (offsetof (struct bp, v) == 0); /* Guaranteed by the C standard */
  b = (struct bp *)breakpoint_list;
  breakpoint_list = NULL;
  breakpoint_list_tail = (struct bp **)&breakpoint_list;
  while (b != NULL)
    {
      struct bp *next;
      
      free (b->v.expr);
      free (b->v.compiled);
      next = (struct bp *)b->v.next;
      free (b);
      b = next;
    }
}
