/* A simple module providing a read-only bit with a constant value */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>

#include <cereal.h>
#include <lib.h>
#include <module.h>

/* One bit-constant instance state */
struct mi
{
  int val;
  struct port_fn write;
};

/* Create a new module instance;
   returns: module instance pointer if OK, NULL on error */
static void *
bit_constant_mi_new (void)
{
  struct mi *mi;

  mi = xmalloc (sizeof (*mi));
  mi->val = 0;
  mi->write.fn.write_1 = NULL;
  return mi;
}

/* Delete MI_ (may be in any emulation state) */
static void
bit_constant_mi_delete (void *mi_)
{
  struct mi *mi;

  mi = mi_;
  free (mi);
}

/* Our XML namespace */
static const char bit_constant_xml_ns[] =
"http://www.fw.cz/mitr/cereal/modules/bit_constant";

/* Save state of MI_ to NODE */
static void
bit_constant_save_state (void *mi_, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;

  mi = mi_;
  ns = xmlNewNs (node, bit_constant_xml_ns, "bit_constant");
  xmlNewTextChild (node, ns, "value", mi->val != 0 ? "1" : "0");
}

/* Load state of MI_ from NODE of DOC;
   returns: 0 if OK, -1 on error */
static int
bit_constant_load_state (void *mi_, xmlDocPtr doc, xmlNodePtr node)
{
  struct mi *mi;
  xmlNsPtr ns;
  int res;

  mi = mi_;
  ns = xmlSearchNsByHref (doc, node, bit_constant_xml_ns);
  res = 0;
  if (ns == NULL)
    goto err;
  for (node = node->children; node != NULL; node = node->next)
    {
      unsigned long val;

      if (node->type != XML_ELEMENT_NODE || node->ns != ns)
	continue;
      if (strcmp (node->name, "value") == 0)
	{
	  res = get_xml_num (doc, node, 2, &val);
	  if (res != 0)
	    goto err;
	  mi->val = val;
	}
    }
 err:
  return res;
}

/* Get number of ports of this space in MI_ */
static size_t
space_get_size (void *mi_)
{
  (void)mi_;
  return 1;
}

/* Read bit port value (for emulation or display; may have no side effects);
   returns: port value */
static int 
space_display_fn (void *mi_, intptr_t data)
{
  struct mi *mi;
  
  (void)data;
  mi = mi_;
  return mi->val != 0;
}

/* Save port_fn for PORT reading to DEST */
static void
space_get_read_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.read_1 = space_display_fn;
  dest->mi = mi_;
}

/* Save port_fn for PORT displaying to DEST */
static void
space_get_display_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.display_1 = space_display_fn;
  dest->mi = mi_;
}

/* Modify bit port value by user */
static void
space_modify_fn (void *mi_, intptr_t data, int val)
{
  struct mi *mi;

  (void)data;
  mi = mi_;
  mi->val = (val != 0);
  if (mi->write.fn.write_1 != NULL)
    mi->write.fn.write_1 (mi->write.mi, mi->write.data, val);
}

/* Save port_fn for PORT modification to DEST */
static void
space_get_modify_fn (void *mi_, size_t port, struct port_fn *dest)
{
  assert (port == 0);
  dest->fn.modify_1 = space_modify_fn;
  dest->mi = mi_;
}

/* Set peer write_fn for PORT in MI_ to SRC;
   returns: 0 if OK, errno otherwise */
static int
space_set_write_fn (void *mi_, size_t port, const struct port_fn *src)
{
  struct mi *mi;

  mi = mi_;
  assert (port == 0);
  if (mi->write.fn.write_1 != NULL && src->fn.read_1 != NULL)
    return EBUSY;
  mi->write = *src;
  return 0;
}

static const struct space bit_constant_space =
  {
    .get_size = space_get_size, .get_fn
    {
      [PORT_TYPE_READ] = space_get_read_fn, [PORT_TYPE_WRITE] = NULL,
      [PORT_TYPE_DISPLAY] = space_get_display_fn,
      [PORT_TYPE_MODIFY] = space_get_modify_fn,
    },
    .set_fn =
    {
      [PORT_TYPE_READ] = NULL, [PORT_TYPE_WRITE] = space_set_write_fn,
      [PORT_TYPE_DISPLAY] = NULL, [PORT_TYPE_MODIFY] = NULL
    }
  };

struct cereal_module bit_constant_module =
  {
    .next = NULL, .pprev = NULL, .name = "bit_constant", .xml = NULL,
    .handle = NULL, .mi_new = bit_constant_mi_new,
    .mi_delete = bit_constant_mi_delete,
    .set_option = generic_module_set_option,
    .get_option = generic_module_get_option,
    .save_setup = generic_module_save_setup,
    .load_setup = generic_module_load_setup,
    .save_state = bit_constant_save_state,
    .load_state = bit_constant_load_state, .w =
    {
      [WIDTH_1] =
      { .spaces = &bit_constant_space, .num_spaces = 1, .xml_spaces = NULL },
      [WIDTH_8] = { .spaces = NULL, .num_spaces = 0, .xml_spaces = NULL },
      [WIDTH_16] = { .spaces = NULL,.num_spaces = 0, .xml_spaces = NULL }
    }
  };
